// This file may be redistributed and modified only under the terms of
// the GNU General Public License (See COPYING for details).
// Copyright (C) 2003 Henri Kuuste

#include "efmath.h"

#define LG_NSEGS 8
#define NSEGS (1<<LG_NSEGS)


void QuatToMatrix(CalQuaternion quat, float m[4][4]) {
  float wx, wy, wz, xx, yy, yz, xy, xz, zz, x2, y2, z2;

  x2 = quat.x + quat.x;
  y2 = quat.y + quat.y;
  z2 = quat.z + quat.z;
  xx = quat.x * x2;
  xy = quat.x * y2;
  xz = quat.x * z2;
  yy = quat.y * y2;
  yz = quat.y * z2;
  zz = quat.z * z2;
  wx = quat.w * x2;
  wy = quat.w * y2;
  wz = quat.w * z2;

  m[0][0] = 1.0 - (yy + zz);
  m[0][1] = xy - wz;
  m[0][2] = xz + wy;
  m[0][3] = 0.0;
             
  m[1][0] = xy + wz;
  m[1][1] = 1.0 - (xx + zz);
  m[1][2] = yz - wx;
  m[1][3] = 0.0;
  
  m[2][0] = xz - wy;
  m[2][1] = yz + wx;  
  m[2][2] = 1.0 - (xx + yy);
  m[2][3] = 0.0;
    
  m[3][0] = 0;
  m[3][1] = 0;
  m[3][2] = 0;
  m[3][3] = 1;  
}

void applyQuaternion(CalQuaternion quat) {
  applyCalMatrix(CalMatrix(quat));
}

void applyCalMatrix(CalMatrix m) {
  float rotation_matrix[4][4];
  CalMatrixToMatrix(m, rotation_matrix);
  glMultMatrixf(&rotation_matrix[0][0]);
}

void CalMatrixToMatrix(CalMatrix cm, float m[4][4]) {
  m[0][0] = cm.dxdx;
  m[1][0] = cm.dxdy;
  m[2][0] = cm.dxdz;
  m[3][0] = 0;
	
  m[0][1] = cm.dydx;
  m[1][1] = cm.dydy;
  m[2][1] = cm.dydz;
  m[3][1] = 0;

  m[0][2] = cm.dzdx;
  m[1][2] = cm.dzdy;
  m[2][2] = cm.dzdz;
  m[3][2] = 0;

  m[0][3] = 0;
  m[1][3] = 0;
  m[2][3] = 0;
  m[3][3] = 1;	
}

void QuatToBallPoints(const CalQuaternion q, CalVector *from, CalVector *to) {
  GLfloat s;
  s = sqrt(q.x*q.x + q.y*q.y);
  if(s == 0.0f) {
    from = new CalVector(0.0f, 1.0f, 0.0f);
  } else {
    from = new CalVector(-q.y/s, q.x/s, 0.0f);
  }
  to->x = q.w*from->x - q.z*from->y;
  to->y = q.w*from->y + q.z*from->x;
  to->z = q.x*from->y - q.y*from->x;
  if(q.w < 0.0f) {
    *from = CalVector(-from->x, -from->y, 0.0);
  }
}

CalQuaternion QuatFromBallPoints(CalVector from, CalVector to)
{
  //TODO look up WFMath::Quaternion::rotation() to see how to do this
  CalQuaternion qu;
  qu.x = from.y*to.z - from.z*to.y;
  qu.y = from.z*to.x - from.x*to.z;
  qu.z = from.x*to.y - from.y*to.x;
  qu.w = from.x*to.x + from.y*to.y + from.z*to.z;
  return qu;
}

/*
  CalVector ConstrainToAxis(CalVector p, CalVector axis) {
  CalVector onPlane;
  GLfloat norm;
  onPlane = p - axis*(axis*p);
  norm = onPlane
  }*/

CalVector Bisect(CalVector v0, CalVector v1)
{
  //TODO look up if CalVectors are normalized already
  CalVector new_vector(0, 0, 0);
  float normalized_vector;
  new_vector = v0+v1;
  normalized_vector = new_vector.x*new_vector.x + new_vector.y*new_vector.y + new_vector.z*new_vector.z;
  if (normalized_vector < 1.0e-5) {
    new_vector = CalVector(0, 0, 1);
  } else {
    new_vector *= (1/sqrt(normalized_vector));
  }
  return (new_vector);
}

void DrawAnyArc(CalVector vFrom, CalVector vTo)
{
  int i;
  CalVector pts[NSEGS+1];
  double dot;
    
  pts[0] = vFrom;
  pts[1] = vTo; pts[NSEGS] = vTo;
  for (i=0; i<LG_NSEGS; i++)
    pts[1] = Bisect(pts[0], pts[1]);
  dot = 2.0*(pts[0]*pts[1]);
  for (i=2; i<NSEGS; i++) {
    pts[i] = (pts[i-1]*dot) - pts[i-2];
  }
  glBegin(GL_LINE_STRIP);
  for (i=0; i<=NSEGS; i++)
    glVertex3f(pts[i].x, pts[i].y, pts[i].z);
  glEnd();
}

void DrawCircle(GLfloat radius) {
  GLfloat a = -0.1;
  glBegin(GL_LINE_STRIP);
  do {
    a += 0.1;
    glVertex3f(radius * cos(a), radius * sin(a), 0);
  } while (a <= 2*PI);
  glEnd();
}

CalQuaternion EulerToQuat(CalVector v) {
  double cr, cp, cy, sr, sp, sy, cpcy, spsy;

  cr = cos(v.x/2);
  cp = cos(v.y/2);
  cy = cos(v.z/2);

  sr = sin(v.x/2);
  sp = sin(v.y/2);
  sy = sin(v.z/2);
	
  cpcy = cp * cy;
  spsy = sp * sy;

  CalQuaternion quat;
  quat.w = cr * cpcy - sr * spsy;
  quat.y = -sr * cpcy - cr * spsy;
  quat.z = -cr * sp * cy + sr * cp * sy;
  quat.w = -cr * cp * sy - sr * sp * cy;

  return quat;
}

CalVector QuatToEuler(CalQuaternion q1) {
  double sqw = q1.w*q1.w;
  double sqx = q1.x*q1.x;
  double sqy = q1.y*q1.y;
  double sqz = q1.z*q1.z;
	
  CalVector v;
  v.z = atan2(-2.0 * (q1.x*q1.y + q1.z*q1.w),(sqx - sqy - sqz + sqw));
  v.y = asin(-2.0 * (q1.x*q1.z - q1.y*q1.w));
  v.x = atan2(2.0 * (q1.y*q1.z + q1.x*q1.w),(-sqx - sqy + sqz + sqw));
  return v;
}

CalQuaternion QuatFromAxisAngle(int axis, double angle) {
  double half_angle = angle / 2;
  CalQuaternion q;
	
  q.w = cos(half_angle);
  for(int i = 0; i < 3; ++i) {
    switch(i) {
    case 0:
      q.x = (i == axis) ? sin(half_angle) : 0;
      break;
    case 1:
      q.y = (i == axis) ? sin(half_angle) : 0;
      break;
    case 2:
      q.z = (i == axis) ? sin(half_angle) : 0;
      break;
    }
  }
  return q;
}
