// This file may be redistributed and modified only under the terms of
// the GNU General Public License (See COPYING for details).
// Copyright (C) 2003 Henri Kuuste

#include "image_loader.h"
#include <fstream>
#include <iostream>
#include <GL/glu.h>
#include <gdkmm/pixbuf.h>
#include <glibmm/refptr.h>

GLuint loadGTK(const std::string &filename) {
  Glib::RefPtr<Gdk::Pixbuf> pb;
  try {
    pb = Gdk::Pixbuf::create_from_file(filename);
  }
  catch (std::exception e) {
    std::cerr << "Texture file '" << filename << "' not loaded." << std::endl;
    std::cerr << "Excepton: " << e.what() << std::endl << std::flush;
    return 0;
  }
  catch (...) {
    std::cerr << "Texture file '" << filename << "' not loaded. Unknown exception." << std::endl;
    
    return 0;
  }
  if(!pb) {
    std::cerr << "Texture file '" << filename << "' not found." << std::endl;
    return 0;
  }

  // Flip the darn thing
  const guint8* pBuffer = pb->get_pixels();
  guint8* tBuffer;
  tBuffer = new unsigned char[pb->get_width() * pb->get_height() * pb->get_n_channels()];
  int y;
  for(y = 0; y < pb->get_height(); y++) {
    memcpy(&tBuffer[y * pb->get_rowstride()],
           &pBuffer[(pb->get_height() - y - 1) * pb->get_rowstride()],
           pb->get_rowstride());
  }
	
  Glib::RefPtr<Gdk::Pixbuf> tb = Gdk::Pixbuf::create_from_data(tBuffer,
                                                               pb->get_colorspace(),
                                                               pb->get_has_alpha(),
                                                               pb->get_bits_per_sample(),
                                                               pb->get_width(),
                                                               pb->get_height(),
                                                               pb->get_rowstride());

  // Load it into OpenGL
  glPushAttrib(GL_TEXTURE_BIT | GL_PIXEL_MODE_BIT);
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

  GLuint textureId;
  glGenTextures(1, &textureId);
  glBindTexture(GL_TEXTURE_2D, textureId);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexImage2D(GL_TEXTURE_2D, 0, tb->get_has_alpha() ? GL_RGBA : GL_RGB,
               tb->get_width(), tb->get_height(),
               0, tb->get_has_alpha() ? GL_RGBA : GL_RGB, GL_UNSIGNED_BYTE,
               tb->get_pixels());

  glPopAttrib();
  return textureId;
}


GLuint loadRAW(const std::string &filename) {
  std::ifstream file;
  file.open(filename.c_str(), std::ios::in | std::ios::binary);
  if(!file) {
    std::cerr << "Texture file '" << filename << "' not found." << std::endl;
    return 0;
  }

  int width;
  file.read((char *)&width, 4);
  int height;
  file.read((char *)&height, 4);
  int depth;
  file.read((char *)&depth, 4);

  unsigned char *pBuffer;
  pBuffer = new unsigned char[2 * width * height * depth];
  if(pBuffer == 0) {
    std::cerr << "Memory allocation for texture '" << filename << "' failed." << std::endl;
    return 0;
  }

  file.read((char *)pBuffer, width * height * depth);

  file.close();

  int y;
  for(y = 0; y < height; y++) {
    memcpy(&pBuffer[(height + y) * width * depth], &pBuffer[(height - y - 1) * width * depth], width * depth);
  }

  glPushAttrib(GL_TEXTURE_BIT | GL_PIXEL_MODE_BIT);
  glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

  GLuint textureId;
  glGenTextures(1, &textureId);
  glBindTexture(GL_TEXTURE_2D, textureId);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  glTexImage2D(GL_TEXTURE_2D, 0, (depth == 3) ? GL_RGB : GL_RGBA,
               width, height,
               0, (depth == 3) ? GL_RGB : GL_RGBA, GL_UNSIGNED_BYTE,
               &pBuffer[width * height * depth]);

  delete [] pBuffer;
  glPopAttrib();
  return textureId;
}
