// This file may be redistributed and modified only under the terms of
// the GNU General Public License (See COPYING for details).
// Copyright (C) 2003 Henri Kuuste
// Changes Copryright 2004 Kai Blin

#include "mainWindow.h"
#include "efmath.h"

std::string duplicateSlashes(const std::string &str) {
  std::string s(str);
  std::string::size_type pos = 0;
  while((pos = s.find_first_of("\\", pos)) != std::string::npos) {
    s = s.replace(pos, 1, "\\\\");
    pos += 4;
  }
  return s;
}

mainWindow::mainWindow() :
  calm(NULL), settingMaterial(false), frames(0), frameSeconds(0),
  m_select_mode(false), selectedBone(0), 
  boneSelected(false), objToRotate(-1), ball(CalVector(0, 0, 0), 0.75),
  aMode(NONE), rotMode(ARCBALL), selectionFlags(SUBMESHES), frontFace(GL_CW),
  lit(false), skeleton(false), wireframe(false), 
  wOnSurface(false), culling(false),
  lastRotX(0), lastRotY(0), lastTranX(0), lastTranY(0),
  initT(0, 0, -270), initR(-70, -30, 0),
  lastPath(""), num_recent_files(6), tMenu(NULL), 
  lastSeconds(0)
{
  selColor[0] = 0.65;
  selColor[1] = 0.24;
  selColor[2] = 0.24;
  createGUI();
  reset_transforms();
  loadConfig();
}

mainWindow::~mainWindow() {
  saveConfig();
  if(aboutDialog) {
    delete aboutDialog;
  }
  if(controlsHelpDialog) {
    delete controlsHelpDialog;
  }
  if(executeActionDialog) {
    delete executeActionDialog;
  }
  if(calm) {
    delete calm;
  }
  if(mainWin) {
    delete mainWin;
  }
  s_objects.clear();
}

void mainWindow::on_new() {
  // IMPLEMENT ME
}

void mainWindow::on_open() {
  bool back = true;
  Gtk::FileSelection dialog("Select Cal3D cfg file to open");
  dialog.set_transient_for(*mainWin);
  dialog.property_window_position().set_value(Gtk::WIN_POS_CENTER_ON_PARENT);
  dialog.set_modal();
  dialog.set_filename(lastPath);
  do {
    back = true;
    int response = dialog.run();
    if(response == Gtk::RESPONSE_OK) {
      if(dialog.get_filename().substr(dialog.get_filename().length()-3) != "cfg") {
        Gtk::MessageDialog notCfg(dialog, "Not a cfg file.\nTry again?", false, 
                                  Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK_CANCEL, true);
        notCfg.set_transient_for(dialog);
        notCfg.property_window_position().set_value(Gtk::WIN_POS_CENTER_ON_PARENT);
        response = notCfg.run();
        if(response == Gtk::RESPONSE_OK) {
          back = false;
        }
      } else {
        if(!loadCFG(dialog.get_filename())) {
          Gtk::MessageDialog badCfg(dialog, "Bad cfg file.\nTry again?", false, 
                                    Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK_CANCEL, true);
          badCfg.set_transient_for(dialog);
          badCfg.property_window_position().set_value(Gtk::WIN_POS_CENTER_ON_PARENT);
          response = badCfg.run();
          if(response == Gtk::RESPONSE_OK) {
            back = false;
          }
        }
      }
    }
  } while (!back);
}

void mainWindow::on_save() {
  // IMPLEMENT ME
}

void mainWindow::on_save_as() {
  // IMPLEMENT ME
}

void mainWindow::on_import(){
  bool back = true;
  Gtk::FileSelection dialog("Select file to import");
  dialog.set_transient_for(*mainWin);
  dialog.property_window_position().set_value(Gtk::WIN_POS_CENTER_ON_PARENT);
  dialog.set_modal();
  dialog.set_filename(lastPath);
  do {
    back = true;
    int response = dialog.run();
    if(response == Gtk::RESPONSE_OK) {

      std::string filename = dialog.get_filename();
      std::string filetype = filename.substr(filename.length()-3);
      if( filetype == "cfg"){
        if(!loadCFG(filename)) {
          Gtk::MessageDialog badCfg(dialog, "Bad cfg file.\nTry again?", false, 
                                    Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK_CANCEL, true);
          badCfg.set_transient_for(dialog);
          badCfg.property_window_position().set_value(Gtk::WIN_POS_CENTER_ON_PARENT);
          response = badCfg.run();
          if(response == Gtk::RESPONSE_OK) {
            back = false;
          }
        }
        else {
          return;
        }
      }
             
      if(filetype == "obj" ){
        std::cout << "Loading .obj file." << std::endl;
        loadObj(filename);
        return;
      }
      
      if(filetype == "md3") {
        std::cout << "Filename = \"" << filename << "\"" << std::endl;
        loadMd3(filename);
        return;
      }
      else{
        Gtk::MessageDialog notCfg(dialog, "Not a supported file format.\nTry again?",false,
                                  Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK_CANCEL, true);
        notCfg.set_transient_for(dialog);
        notCfg.property_window_position().set_value(Gtk::WIN_POS_CENTER_ON_PARENT);
        response = notCfg.run();
        if(response == Gtk::RESPONSE_OK) {
          back = false;
        }
      }
    }
  } while (!back);
}

void mainWindow::refresh() {
  loadCFG(lastPath);
}


void mainWindow::reset_transforms() {
  t = initT;
  ball.setState(CalVector(-85, -25, 0));
  r = initR;
}

void mainWindow::set_rotMode(RotationMode m, bool elsewhere) {
  if(m == ARCBALL) {
    if(rotMode == EULER) {
      reset_transforms();
      //			ball.setState(r);
    }
    rotMode = ARCBALL;
    if(elsewhere) {
      if(!mode_arcball->get_active()) { mode_arcball->set_active(); }
    }
  } else if(m == EULER) {
    if(rotMode == ARCBALL) {
      reset_transforms();
      //			r = ball.getState();
    }
    rotMode = EULER;
    if(elsewhere) {
      if(!mode_euler->get_active()) { mode_euler->set_active(); }
    }
  }
}

void mainWindow::rotate_scene() {
  switch(rotMode) {
  case ARCBALL:
    ball.update();
    applyCalMatrix(ball.value());
    for(unsigned int i = 0; i < s_objects.size(); i++) {
      s_objects[i]->adjustRotation(ball.qValue());
    }
    break;
  case EULER:
    glRotatef(r.x, 1.0f, 0.0f, 0.0f);
    glRotatef(r.y, 0.0f, 0.0f, 1.0f);
    glRotatef(r.z, 0.0f, 1.0f, 0.0f);
    ball.setState(r);
    for(unsigned int i = 0; i < s_objects.size(); i++) {
      s_objects[i]->adjustRotation(ball.qValue());
    }
  }
}

void mainWindow::renderGL(double seconds) {
  frames++;
  if((seconds - frameSeconds) >= 1) {
    std::ostringstream s;
    s << (int)(frames/(seconds-frameSeconds));
    fps->set_text(s.str());
    frameSeconds = seconds;
    frames = 0;
  }
  if(calm) {
    calm->update((seconds - lastSeconds) * timeScale->get_value());
    lastSeconds = seconds;
    glPushAttrib( GL_ALL_ATTRIB_BITS );
    if(lit) {
      glEnable( GL_LIGHTING );
    } else {
      glDisable( GL_LIGHTING);
    }
    if(culling) {
      glFrontFace(frontFace);
      glCullFace( GL_BACK );
      glEnable( GL_CULL_FACE );
    }
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glTranslatef(t.x * calm->scale, t.y * calm->scale, t.z * calm->scale);
    rotate_scene();
    glTranslatef(0.0f, 0.0f, -90.f * calm->scale);
    if(skeleton) {
      calm->renderSkeleton();
    }
    if(!skeleton || (skeleton && wireframe)) {
      calm->render();
    }
    glPopAttrib();
    for(unsigned int i = 0; i < s_objects.size(); i++) {
      s_objects[i]->render();
    }
    if((aMode == ROTATING_OBJECTS) && (objToRotate >= 0)) {
      s_objects[objToRotate]->ball.render();
    }
  }
}

void mainWindow::getSelection(double x, double y) {
	
  fprintf(stderr, "In getSelection %f, %f\n", x, y);
  if(calm) {
    assert(viewport);
    Glib::RefPtr<Gdk::GL::Window> glwindow = viewport->get_gl_window();
    assert(glwindow);
    if (!glwindow->gl_begin(viewport->get_gl_context()))
      return;
    glPushAttrib(GL_ALL_ATTRIB_BITS);
    GLuint buffer[HITS_BUFFER_SIZE];
    GLint hits;
    GLint viewport_v[4];
    glGetIntegerv(GL_VIEWPORT, viewport_v);
    glSelectBuffer(HITS_BUFFER_SIZE, buffer);
    (void)glRenderMode(GL_SELECT);
    glInitNames();
    glPushName(0);
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    if(skeleton && !wireframe) {
      gluPickMatrix((GLdouble)x, (GLdouble) (viewport_v[3]-y), 12.0, 12.0, viewport_v);
    } else {
      gluPickMatrix((GLdouble)x, (GLdouble) (viewport_v[3]-y), 6.0, 6.0, viewport_v);
    }
			
    viewport->doPerspective();

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glTranslatef(t.x * calm->scale, t.y * calm->scale, t.z * calm->scale);
    rotate_scene();
    glTranslatef(0.0f, 0.0f, -90.f * calm->scale);
    if(skeleton && !wireframe && selectionFlags != STATIC_OBJECTS) {
      calm->renderSkeleton(true);
    } else if(selectionFlags == SUBMESHES) {
      calm->render(true);
    } else if(selectionFlags == STATIC_OBJECTS) {
      for(unsigned int i = 0; i < s_objects.size(); i++) {
        glLoadName(i);
        s_objects[i]->render();
      }
    }
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glFlush();
    hits = glRenderMode(GL_RENDER);
    glPopAttrib();
    glwindow->gl_end();
    processHits(hits, buffer);
  }
}

void mainWindow::processHits(GLint hits, GLuint buffer[]) {
  GLuint meshId=0, submeshId=0, names, *ptr;
  GLuint nearest = UINT_MAX;
  ptr = (GLuint *)buffer;
  for(GLint i = 0; i < hits; i++) {
    names = *ptr;
    ptr++;
    if(*ptr < nearest) {
      nearest = *ptr;
      ptr += 2;
      for(GLuint j = 0; j < names; j++) {
        if(j == 0) {
          meshId = *ptr;
        } else {
          submeshId = *ptr;
        }
        ptr++;
      }
    } else {
      ptr += names + 2;
    }
  }
  if(hits > 0 && calm) {
    if(skeleton && !wireframe && selectionFlags != STATIC_OBJECTS) {
      boneSelected = true;
      calm->boneSelected = true;
      selectedBone = meshId;
      calm->selectedBone = meshId;
    } else if(selectionFlags == SUBMESHES) {
      Gtk::TreeModel::Row row = calm->meshes.children()[meshId].children()[submeshId];
      entityTree->expand_to_path(Gtk::TreeModel::Path(row));
      assert(refTreeSelection);
      refTreeSelection->select(row);
      //Gtk::TreeModel::iterator iter = refTreeSelection->get_selected();
    } else if(selectionFlags == STATIC_OBJECTS) {
      selectionFlags = SUBMESHES;
      if(aMode == ROTATING_OBJECTS) {
        helpBar->push("LMB to rotate, RMB to end rotation mode");
        objToRotate = meshId;
      } else {
        helpBar->push("");
        s_objects[meshId]->setup();
      }
    }
  }
}

void mainWindow::on_recent(Glib::ustring filename) {
  if(!loadCFG(filename)) {
    Gtk::MessageDialog badCfg(*mainWin, "Bad cfg file.",false,
                              Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK, true);
    badCfg.set_transient_for(*mainWin);
    badCfg.property_window_position().set_value(Gtk::WIN_POS_CENTER_ON_PARENT);
    badCfg.run();
  }
}

bool mainWindow::loadCFG(const Glib::ustring filename) {
  Model *tCalm = new Model();
  lastPath = filename;
  if(!tCalm->loadCFG(lastPath)) {
    delete tCalm;
    return false;
  } else {
    add_recent_file(filename);
    if(calm) {
      std::cout << "Starting to delete existing data..." << std::endl;
      entityTree->remove_all_columns();
      if(!infoBox->children().empty()) {
        infoBox->children().clear();
      }
      delete calm;
      boneSelected = false;

      lastSelected = Gtk::TreeIter();
      s_objects.clear();
    }
    calm = tCalm;
    tree_from_model();
    viewport->timer.reset();
    viewport->setPlanes(30.0f * calm->scale, 5000.0f * calm->scale);
    lastSeconds = 0;
    frameSeconds = 0;
    frames = 0;
    calm->light = lit;
    calm->wireframe = wireframe;
    calm->surface = wOnSurface;
    float temp = lodScale->get_value() / 100;
    calm->setLodLevel(temp);
    calm->setSelColor(selColor);
    setPolyCount();
  }
  return true;
}

bool mainWindow::loadMd3(const std::string filename) {
  Md3Model *object =  new Md3Model();
  if(!object->readFile(filename)){
    delete object;
    return false;
  }
  else {
    object->render();
  }
  return true;
}

bool mainWindow::loadObj(const std::string filename) {
  Obj *object = new Obj(filename);
  if(!object){
    std::cerr << "loading OBJ failed." << std::endl;
    return false;
  }
  else {
    object->render();
  }
  return true;
}

void mainWindow::add_recent_file(const Glib::ustring &filename) {
  recentFiles.remove(filename);
  recentFiles.push_front(filename);
  if(recentFiles.size() > num_recent_files) {
    recentFiles.pop_back();
  }
  if(recentFilesMenu) {
    recentFilesMenu->items().clear();
    std::list<Glib::ustring>::iterator it;
    Gtk::MenuItem *temp;
    for(it = recentFiles.begin(); it != recentFiles.end(); it++) {
      temp = new Gtk::MenuItem(*it);
      temp->signal_activate().connect(
                                      sigc::bind(
                                                 sigc::mem_fun(*this, &mainWindow::on_recent),
                                                 *it));
      recentFilesMenu->append(*Gtk::manage(temp));
    }
    recentFilesMenu->show_all();
  }
}


void mainWindow::on_cal3d_cfg_export() {
  // IMPLEMENT ME
}

void mainWindow::on_menu_cw() {
  if(cw->get_active()) {
    frontFace = GL_CW;
  } else {
    frontFace = GL_CCW;
  }
}

void mainWindow::on_menu_ccw() {
  if(ccw->get_active()) {
    frontFace = GL_CCW;
  } else {
    frontFace = GL_CW;
  }
}

void mainWindow::on_static_object() {
  if(boneSelected) {
    StaticObject *o = new StaticObject();
    o->createSetupDialog(mainWin);
    o->attach(calm->calModel->getSkeleton()->getBone(selectedBone));
    s_objects.push_back(o);
    helpBar->push("");
  } else {
    helpBar->push("You have to have a bone selected!");
  }
}

void mainWindow::on_obj() {
  if(boneSelected) {
    Obj *o = new Obj();
    bool back = true;
    Gtk::FileSelection dialog("Select .obj file to import");
    dialog.set_transient_for(*mainWin);
    dialog.property_window_position().set_value(Gtk::WIN_POS_CENTER_ON_PARENT);
    dialog.set_modal();
    do {
      back = true;
      int response = dialog.run();
      if(response == Gtk::RESPONSE_OK) {
        if(dialog.get_filename().substr(dialog.get_filename().length()-3) != "obj") {
          Gtk::MessageDialog notObj(dialog, "Not a obj file.\nTry again?",false,
                                    Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK_CANCEL, true);
          notObj.set_transient_for(dialog);
          notObj.property_window_position().set_value(Gtk::WIN_POS_CENTER_ON_PARENT);
          response = notObj.run();
          if(response == Gtk::RESPONSE_OK) {
            back = false;
          }
        } else {
          if(!o->readFile(dialog.get_filename(), true)) {
            Gtk::MessageDialog badObj(dialog, "Bad obj file.\nTry again?",false,
                                      Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK_CANCEL, true);
            badObj.set_transient_for(dialog);
            badObj.property_window_position().set_value(Gtk::WIN_POS_CENTER_ON_PARENT);
            response = badObj.run();
            if(response == Gtk::RESPONSE_OK) {
              back = false;
            }
          } else {
            o->readFile(dialog.get_filename());
            o->createSetupDialog(mainWin);
            o->attach(calm->calModel->getSkeleton()->getBone(selectedBone));
            s_objects.push_back((StaticObject*)o);
            helpBar->push("");
          }
        }
      }
    } while (!back);
  } else {
    helpBar->push("You have to have a bone selected!");
  }
}

void mainWindow::on_md3() {
  if(boneSelected) {
    Md3Model *object = new Md3Model();
    bool back = true;
    Gtk::FileSelection dialog("Select .md3 file to import");
    dialog.set_transient_for(*mainWin);
    dialog.property_window_position().set_value(Gtk::WIN_POS_CENTER_ON_PARENT);
    dialog.set_modal();
    do {
      back = true;
      int response = dialog.run();
      if(response == Gtk::RESPONSE_OK) {
        if(dialog.get_filename().substr(dialog.get_filename().length()-3) != "md3") {
          Gtk::MessageDialog notObj(dialog, "Not an md3 file.\nTry again?",false,
                                    Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK_CANCEL, true);
          notObj.set_transient_for(dialog);
          notObj.property_window_position().set_value(Gtk::WIN_POS_CENTER_ON_PARENT);
          response = notObj.run();
          if(response == Gtk::RESPONSE_OK) {
            back = false;
          }
        } else {
          if(!object->readFile(dialog.get_filename())) {
            Gtk::MessageDialog badObj(dialog, "Bad md3 file.\nTry again?",false,
                                      Gtk::MESSAGE_ERROR, Gtk::BUTTONS_OK_CANCEL, true);
            badObj.set_transient_for(dialog);
            badObj.property_window_position().set_value(Gtk::WIN_POS_CENTER_ON_PARENT);
            response = badObj.run();
            if(response == Gtk::RESPONSE_OK) {
              back = false;
            }
          } else {
            object->createSetupDialog(mainWin);
            object->attach(calm->calModel->getSkeleton()->getBone(selectedBone));
            s_objects.push_back((StaticObject*)object);
            helpBar->push("");
          }
        }
      }
    } while (!back);
  } else {
    helpBar->push("You have to have a bone selected!");
  }
}

void mainWindow::on_setup_static_object() {
  selectionFlags = STATIC_OBJECTS;
  helpBar->push("Select a static object (Shift+LMB)");
}

void mainWindow::on_rotate_static_objects() {
  selectionFlags = STATIC_OBJECTS;
  aMode = ROTATING_OBJECTS;
  helpBar->push("Select a static object to rotate (Shift+LMB)");
}
	
void mainWindow::on_lod() {
  if(calm) {
    float temp = lodScale->get_value() / 100;
    calm->setLodLevel(temp);
    setPolyCount();
    assert(refTreeSelection);
    Gtk::TreeModel::iterator iter = refTreeSelection->get_selected();
    if(iter && pcLabel) {
      EntityItemType type = (*iter)[cols.type];
      std::ostringstream str;
      switch(type) {
      case MESH:
        str << calm->getMeshPolyCount((*iter)[cols.id]);
        pcLabel->set_text(str.str());
        break;
      case MODEL:
        str << calm->getPolyCount();
        pcLabel->set_text(str.str());
        break;
      case SUBMESH:
        str << calm->getSubmeshPolyCount((*(*iter).parent())[cols.id],
                                         (*iter)[cols.id]);
        pcLabel->set_text(str.str());
        break;
      default:
        break;
      }
    }
    viewport->update();
  }
}

void mainWindow::on_sizeScale() {
  if(calm){
    calm->setScale(pow(10,size_scale->get_value()));
  }
}

void mainWindow::on_lights() {
  lit = !lit;
  if(calm) {
    calm->light = lit;
  }
  std::cout << "Lights " << (lit ? "on" : "off") << std::endl;
}

void mainWindow::on_select() {
  m_select_mode = !m_select_mode;
  std::cout << "Select mode "<<  (m_select_mode ? "on" : "off") << std::endl;
}

void mainWindow::on_skeleton() {
  skeleton = !skeleton;
  std::cout << "Rendering " << (skeleton ? "skeleton" : "mesh") << std::endl;
}

void mainWindow::on_wireframe() {
  wireframe = !wireframe;
  if(calm) {
    calm->wireframe = wireframe;
  }
  std::cout << "Wireframe rendering " << (wireframe ? "on" : "off") << std::endl;
}

void mainWindow::on_backface_culling() {
  culling = backfaceCulling->get_active();
}

void mainWindow::on_wire_on_surface() {
  wOnSurface = wireOnSurface->get_active();
  if(calm) {
    calm->surface = wOnSurface;
  }
}

void mainWindow::on_background_color() {
  Gtk::ColorSelectionDialog dialog;
  dialog.set_transient_for(*mainWin);
  dialog.property_window_position().set_value(Gtk::WIN_POS_CENTER_ON_PARENT);
  dialog.set_modal();
  Gtk::ColorSelection *colorSel = dialog.get_colorsel();
  Gdk::Color current;
  double bg[3];
  viewport->getBackgroundColor(bg);
  current.set_rgb_p(bg[0], bg[1], bg[2]);
  colorSel->set_current_color(current);
  int response = dialog.run();
  if(response == Gtk::RESPONSE_OK) {
    viewport->setBackgroundColor(colorSel->get_current_color().get_red_p(),
                                 colorSel->get_current_color().get_green_p(),
                                 colorSel->get_current_color().get_blue_p());
  }
}

void mainWindow::on_selection_color() {
  Gtk::ColorSelectionDialog dialog;
  dialog.set_transient_for(*mainWin);
  dialog.property_window_position().set_value(Gtk::WIN_POS_CENTER_ON_PARENT);
  dialog.set_modal();
  Gtk::ColorSelection *colorSel = dialog.get_colorsel();
  Gdk::Color current;
  current.set_rgb_p(selColor[0], selColor[1], selColor[2]);
  colorSel->set_current_color(current);
  int response = dialog.run();
  if(response == Gtk::RESPONSE_OK) {
    selColor[0] = colorSel->get_current_color().get_red_p();
    selColor[1] = colorSel->get_current_color().get_green_p();
    selColor[2] = colorSel->get_current_color().get_blue_p();
    if(calm) { 
      calm->setSelColor(selColor);
    }
  }
}

bool mainWindow::on_tree_event(GdkEvent *event) {
  if(event->type == GDK_BUTTON_PRESS) {
    if(entityTree->get_model()) {
      assert(refTreeSelection);
      Gtk::TreeModel::iterator iter = refTreeSelection->get_selected();
      if(event->button.button == 3) {
        if(tMenu) {
          delete tMenu;
        }
        pcLabel = NULL;
        tMenu = Gtk::manage(new Gtk::Menu);
        Gtk::Menu::MenuList &list = tMenu->items();
        if(iter) {
          Gtk::TreeModel::Row row = *iter;
          switch(row[cols.type]) {
          case ANIMATION_ACTION:
            list.push_back(Gtk::Menu_Helpers::MenuElem("Execute",
                                                       sigc::bind(
                                                                  sigc::mem_fun(*this,
                                                                             &mainWindow::execute_action_handler),
                                                                  true)));
            list.push_back(Gtk::Menu_Helpers::MenuElem("Fast Execute",
                                                       sigc::bind(
                                                                  sigc::mem_fun(*this,
                                                                             &mainWindow::execute_action_handler),
                                                                  false)));
            tMenu->popup(event->button.button, event->button.time);
            break;
          case SUBMESH:
            list.push_back(Gtk::Menu_Helpers::MenuElem("Change Material",
                                                       sigc::mem_fun(*this,
                                                                  &mainWindow::change_material_handler)));
            tMenu->popup(event->button.button, event->button.time);
            break;
          case ANIMATION_CYCLE:
            list.push_back(Gtk::Menu_Helpers::MenuElem("Smoothly Blend into",
                                                       sigc::mem_fun(*this,
                                                                  &mainWindow::blend_in_cycle)));
            tMenu->popup(event->button.button, event->button.time);
          default:
            break;
          }
        }
      }
    }
  } else if (event->type == GDK_2BUTTON_PRESS) {
    if(entityTree->get_model()) {
      assert(refTreeSelection);
      if(event->button.button == 1) {

        Gtk::TreeModel::iterator iter = refTreeSelection->get_selected();
        if(iter) {
          Gtk::TreeModel::Row row = *iter;
          switch(row[cols.type]) {
          case ANIMATION_ACTION:
            execute_action_handler(false);
            break;
          case ANIMATION_CYCLE:
            blend_in_cycle();
            break;
          default:
            break;
          }
        }
      }
    }
  }
  return false;
}

void mainWindow::execute_action_handler(bool read) {
  int retval = 0;
  if(read) {
    retval = executeActionDialog->run();
  }
  if(!read || retval == Gtk::RESPONSE_OK) {
    assert(refTreeSelection);
    Gtk::TreeModel::iterator iter = refTreeSelection->get_selected();
    if(iter) {
      Gtk::TreeModel::Row row = *iter;
      float fadeIn = fadeInDelaySpin->get_value();
      float fadeOut =  fadeOutDelaySpin->get_value();
      calm->executeAction(row[cols.id], fadeIn, fadeOut);
    }
  }
}

void mainWindow::blend_in_cycle() {
  assert(refTreeSelection);
  Gtk::TreeModel::iterator iter = refTreeSelection->get_selected();
  if(iter) {
    Gtk::TreeModel::Row row = *iter;
    if(row[cols.type] == ANIMATION_CYCLE) {
      row[cols.active] = true;
      calm->enableCycle(row[cols.id]);
      calm->animCycles[row[cols.id]]->setBlending(1.0);
    } else {
      return;
    }
  } else {
    return;
  }
}

void mainWindow::change_material_handler() {
  assert(refTreeSelection);
  Gtk::TreeModel::iterator iter = refTreeSelection->get_selected();
  if(iter) {
    Gtk::TreeModel::Row row = *iter;
    if(row[cols.type] == SUBMESH) {
      mSetMeshId = (*row.parent())[cols.id];
      mSetId = row[cols.id];
    } else {
      return;
    }
  } else {
    return;
  }

  settingMaterial = true;
  helpBar->push("Select a material from the tree");
}

void mainWindow::on_tree_selection_changed() {
  Gtk::Box::BoxList children = infoBox->children();
  if(!children.empty()) {
    children.clear();
  }

  assert(refTreeSelection);
  Gtk::TreeModel::iterator iter = refTreeSelection->get_selected();
  if(iter) {
    Gtk::TreeModel::Row row = *iter;	
    if(lastSelected) {
      assert(*lastSelected);
      (*lastSelected)[cols.selected] = false;
    }
    lastSelected = iter;
    row[cols.selected] = true;
    int id = row[cols.id];
    switch(row[cols.type]) {
    case ANIMATION_CYCLE:
      infoBox->pack_start(*Gtk::manage(createCycleInfoTable(iter)), Gtk::PACK_SHRINK);
      infoBox->show_all();
      break;
    case ANIMATION_ACTION:
      break;
    case MESH:
      infoBox->pack_start(*Gtk::manage(createMeshInfoTable(iter)), Gtk::PACK_SHRINK);
      infoBox->show_all();
      break;
    case SUBMESH:
      infoBox->pack_start(*Gtk::manage(createSubmeshInfoTable(iter)), Gtk::PACK_SHRINK);
      infoBox->show_all();
      break;
    case MATERIAL:
      if(settingMaterial) {
        calm->setMaterial(mSetMeshId, mSetId, id);
        helpBar->push("");
        settingMaterial = false;
      }
      infoBox->pack_start(*Gtk::manage(createMaterialInfoTable(iter)), Gtk::PACK_SHRINK);
      infoBox->show_all();
      break;
    case CYCLES:
      infoBox->pack_start(*Gtk::manage(createCyclesInfoTable(iter)), Gtk::PACK_SHRINK);
      infoBox->show_all();
      break;
    case MODEL:
      infoBox->pack_start(*Gtk::manage(createModelInfoTable(iter)), Gtk::PACK_SHRINK);
      infoBox->show_all();
      break;
    default:
      break;
    }
  }
}

void mainWindow::on_cycle_toggle(Gtk::TreeModel::Children::iterator iter) {
  Gtk::TreePath path(iter);
  on_tree_active_toggled(path.to_string());

}

void mainWindow::on_tree_active_toggled(const Glib::ustring &path_string) {
  const Gtk::TreePath path(path_string);
  Gtk::TreeRow row = *(calm->tree->get_iter(path));
  row[cols.active] = !row[cols.active];
  switch(row[cols.type]) {
  case ANIMATION_CYCLE:
    if(row[cols.active]) {
      calm->enableCycle(row[cols.id]);
    } else {
      calm->disableCycle(row[cols.id]);
    }
  default:
    break;
  }
}

CalVector mainWindow::get_mouse_pos(double x, double y) {
  CalVector mousePos((x * viewport->adjW) - 1,1 - (y * viewport->adjH), 0);
  mousePos.x = mousePos.x < -1 ? -1 : mousePos.x; 
  mousePos.x = mousePos.x > 1 ? 1 : mousePos.x; 
  mousePos.y = mousePos.y < -1 ? -1 : mousePos.y; 
  mousePos.y = mousePos.y > 1 ? 1 : mousePos.y;
  return mousePos;
}

bool mainWindow::on_viewport_button_press(GdkEventButton *event) {
  if(viewport) {
    CalVector mousePos = get_mouse_pos(event->x, event->y);
    ball.mouse(mousePos);
    for(unsigned int i = 0; i < s_objects.size(); i++) {
      s_objects[i]->ball.mouse(mousePos);
    }
  }
  if(event->button == 1) {
    //Note that everywhere I check whether the shift etc is held
    //down.  Not whether it equals SHIFT, because it seems
    //event->state contains other modifiers as well.
    if(m_select_mode) {
      getSelection(event->x, event->y);	
    } else if(/*(event->state & GDK_BUTTON1_MASK) &&*/ 
              ((aMode == ROTATING_OBJECTS) && (objToRotate >= 0))) {
      //I don't know why it is saying event->state is not GDK_BUTTON1_MASK.  So I commented out this condition, and is seems to work fine anyway, so.. ;)
      s_objects[objToRotate]->ball.beginDrag();
    }
  } else if (event->button == 2) {
    if(event->state & (GDK_BUTTON2_MASK)) {
      ball.beginDrag();
    }
    lastRotX = event->x;
    lastRotY = event->y;
    lastTranX = event->x;
    lastTranY = event->y;
  } else if (event->button == 3) {
    if(aMode == ROTATING_OBJECTS) {
      aMode = NONE;
      objToRotate = -1;
      helpBar->push("");
      return true;
    } else {
      if(tMenu) {
        delete tMenu;
      }
      tMenu = Gtk::manage(new Gtk::Menu);
      Gtk::Menu::MenuList &list = tMenu->items();
      if(entityTree->get_model()) {
        assert(refTreeSelection);
        Gtk::TreeModel::iterator iter = refTreeSelection->get_selected();
        if(iter) {
          Gtk::TreeModel::Row row = *iter;
          switch(row[cols.type]) {
          case SUBMESH:
            list.push_back(Gtk::Menu_Helpers::MenuElem("Change Material",
                                                       sigc::mem_fun(*this,
                                                                  &mainWindow::change_material_handler)));
            break;
          default:
            break;
          }
          list.push_back(Gtk::Menu_Helpers::SeparatorElem());
        }
      }
      if(s_objects.size() > 0) {
        list.push_back(Gtk::Menu_Helpers::MenuElem("Setup Static Object",
                                                   sigc::mem_fun(*this,
                                                              &mainWindow::on_setup_static_object)));
			
        list.push_back(Gtk::Menu_Helpers::MenuElem("Rotate Static Objects",
                                                   sigc::mem_fun(*this,
                                                              &mainWindow::on_rotate_static_objects)));
      }

      list.push_back(Gtk::Menu_Helpers::MenuElem("Reset Transformations",
                                                 sigc::mem_fun(*this,
                                                            &mainWindow::reset_transforms)));

      Gtk::MenuItem *ins = new Gtk::MenuItem("Insert");
      tMenu->append(*Gtk::manage(ins));
      Gtk::Menu *subMenu = new Gtk::Menu();
      ins->set_submenu(*Gtk::manage(subMenu));
	
      Gtk::MenuItem *pMenuItem = new Gtk::MenuItem("Standard static object");
      pMenuItem->signal_activate().connect(
                                           sigc::mem_fun(*this, &mainWindow::on_static_object));
      subMenu->append(*Gtk::manage(pMenuItem));

      pMenuItem = new Gtk::MenuItem(".obj");
      pMenuItem->signal_activate().connect(
                                           sigc::mem_fun(*this, &mainWindow::on_obj));
      subMenu->append(*Gtk::manage(pMenuItem));
	
      Gtk::MenuItem *pMI = new Gtk::MenuItem("Preferences");
      tMenu->append(*Gtk::manage(pMI));
      Gtk::Menu *pM = new Gtk::Menu();
      pMI->set_submenu(*Gtk::manage(pM));
      Gtk::Menu::MenuList &subList = pM->items();
      subList.push_back(Gtk::Menu_Helpers::MenuElem("Background Color",
                                                    sigc::mem_fun(*this,
                                                               &mainWindow::on_background_color)));
      subList.push_back(Gtk::Menu_Helpers::MenuElem("Selection Color",
                                                    sigc::mem_fun(*this,
                                                               &mainWindow::on_selection_color)));
      tMenu->show_all();
      tMenu->popup(event->button, event->time);
    }
  }
  return true;
}

bool mainWindow::on_viewport_button_release(GdkEventButton *event) {
  CalVector v;
  switch (event->button) {
  case 1:
    if(objToRotate != -1) {
      s_objects[objToRotate]->ball.endDrag();
    }
    break;
  case 2:
    ball.endDrag();
    break;
  case 3:
    break;
  default:
    return false;
  }
  return true;
}

bool mainWindow::on_viewport_mouse_move(GdkEventMotion *event) {
  if(event->state & GDK_BUTTON2_MASK && 
     !(event->state & GDK_CONTROL_MASK) &&
     !(event->state & GDK_SHIFT_MASK)) {
    if(rotMode == EULER) {
      r.y += event->x - lastRotX;
      r.x += event->y - lastRotY;
    }
  } else if(event->state & (GDK_BUTTON2_MASK | GDK_CONTROL_MASK) &&
            !(event->state & GDK_SHIFT_MASK)) {
    t.z += (event->y - lastTranY);
  } else if(event->state & (GDK_BUTTON2_MASK | GDK_SHIFT_MASK) &&
            !(event->state & GDK_CONTROL_MASK)) {
    t.x += (event->x - lastTranX);
    t.y -= (event->y - lastTranY);
  }
  lastRotX = event->x;
  lastRotY = event->y;
  lastTranX = event->x;
  lastTranY = event->y;
  CalVector mousePos = get_mouse_pos(event->x, event->y);
  ball.mouse(mousePos);
  for(unsigned int i = 0; i < s_objects.size(); i++) {
    s_objects[i]->ball.mouse(mousePos);
  }
  return true;
}

bool mainWindow::on_main_mouse_move(GdkEventMotion *event) {
  return true;
}

void mainWindow::on_main_key_press(GdkEventKey *event) {
  std::cerr << "In on_main_key_press: event->keyval = " << event->keyval <<"." << std::endl;
  switch(event->keyval) {
  case 65505: //left shift
    // fall through to
  case 65506: { //right shift
    assert(select_button);
    select_button->set_active(true);
    break;
  }
    default:
      return;
  }
}

void mainWindow::on_main_key_release(GdkEventKey *event) {
  std::cerr << "In on_main_key_release: event->keyval = " << event->keyval <<"." << std::endl;
  switch(event->keyval){
  case 65505: //left shift
    // fall through
  case 65506:{ //right shift
    assert(select_button);
    select_button->set_active(false);
    break;
  }
  default:
    return;
  }
}

void mainWindow::tree_from_model() {
  entityTree->append_column("Entity Tree", cols.text);

  Gtk::CellRendererToggle *pTCR = new Gtk::CellRendererToggle();
  Gtk::TreeViewColumn *col = new Gtk::TreeViewColumn("Active", *Gtk::manage(pTCR));
  entityTree->append_column(*Gtk::manage(col));
  col->add_attribute(pTCR->property_active(), cols.active);
  pTCR->property_activatable().set_value(true);
  pTCR->property_xalign().set_value(0.0);
  pTCR->signal_toggled().connect(sigc::mem_fun(*this, &mainWindow::on_tree_active_toggled));	
  entityTree->set_model(calm->tree);
  entityTree->show_all();

  //	entityTree->expand_all();
}

int mainWindow::getPolyCount() {
  if(calm) { 
    return calm->getPolyCount();
  } else {
    return 0;
  }
}

void mainWindow::setPolyCount() {
  if(polyCount) {
    std::ostringstream str;
    str << getPolyCount();
    polyCount->set_text(str.str());
  }
}

void mainWindow::loadConfig() {
  if(conf.readFromFile(getConfigFile())) {
    varconf::Variable temp;
    temp = conf.getItem("layout", "tree_pane");
    if(temp.is_int() && treePane) {
      treePane->set_position((int)temp);
    }
    temp = conf.getItem("layout", "info_pane");
    if(temp.is_int() && infoPane) {
      infoPane->set_position((int)temp);
    }
    if(mainWin) {
      int w = 800, h = 600;
      temp = conf.getItem("layout", "mainwin_width");
      if(temp.is_int()) {
        w = (int)temp;
      }
      temp = conf.getItem("layout", "mainwin_height");
      if(temp.is_int()) {
        h = (int)temp;
      }
      mainWin->resize(w, h);
    }
    temp = conf.getItem("recent_files", "max_num_recent_files");
    if(temp.is_int()) {
      num_recent_files = (int)temp;
    }
    std::list<Glib::ustring> tlist;
    for(unsigned int i = 0; i < num_recent_files; i++) {
      std::ostringstream s;
      s << "recent_file_" << i;
      temp = conf.getItem("recent_files", s.str());
      if(temp.is_string()) {
        tlist.push_front((std::string)temp);
      } else {
        conf.erase("recent_files", s.str());
        break;
      }
    }
    std::list<Glib::ustring>::iterator it;
    for(it = tlist.begin(); it != tlist.end(); it++) {
      add_recent_file(*it);
    }
    if(viewport) {
      double bgColor[3];
      viewport->getBackgroundColor(bgColor);
      temp = conf.getItem("preferences", "bgcolor_red");
      if(temp.is_double()) {
        bgColor[0] = (double)temp;
      }
      temp = conf.getItem("preferences", "bgcolor_green");
      if(temp.is_double()) {
        bgColor[1] = (double)temp;
      }
      temp = conf.getItem("preferences", "bgcolor_blue");
      if(temp.is_double()) {
        bgColor[2] = (double)temp;
      }
      viewport->setBackgroundColor(bgColor[0], bgColor[1], bgColor[2]);
    }
    temp = conf.getItem("preferences", "selcolor_red");
    if(temp.is_double()) {
      selColor[0] = (double)temp;
    }
    temp = conf.getItem("preferences", "selcolor_green");
    if(temp.is_double()) {
      selColor[1] = (double)temp;
    }
    temp = conf.getItem("preferences", "selcolor_blue");
    if(temp.is_double()) {
      selColor[2] = (double)temp;
    }
    if(calm) {
      calm->setSelColor(selColor);
    }
    temp = conf.getItem("preferences", "rotation_mode");
    if(temp.is_int()) {
      int t = (int)temp;
      set_rotMode((RotationMode)t, true);
    }
  }
}

void mainWindow::saveConfig() {
  varconf::Variable temp;
  if(treePane) {
    temp = treePane->get_position();
    conf.setItem("layout", "tree_pane", temp);
  }
  if(infoPane) {
    temp = infoPane->get_position();
    conf.setItem("layout", "info_pane", temp);
  }
  if(mainWin) {
    temp = mainWin->get_width();
    conf.setItem("layout", "mainwin_width", temp);
    temp = mainWin->get_height();
    conf.setItem("layout", "mainwin_height", temp);
  }
  if(viewport) {
    double bgColor[3] = {0, 0, 0};
    viewport->getBackgroundColor(bgColor);
    temp = bgColor[0];
    conf.setItem("preferences", "bgcolor_red", temp);
    temp = bgColor[1];
    conf.setItem("preferences", "bgcolor_green", temp);
    temp = bgColor[2];
    conf.setItem("preferences", "bgcolor_blue", temp);
  }
  temp = rotMode;
  conf.setItem("preferences", "rotation_mode", temp);
  std::list<Glib::ustring>::iterator it;
  int i = 0;
  for(it = recentFiles.begin(); it != recentFiles.end(); it++, i++) {
    temp = duplicateSlashes((std::string)*it);
    std::ostringstream s;
    s << "recent_file_" << i;
    conf.setItem("recent_files", s.str(), temp);
  }
  temp = (int) num_recent_files;
  conf.setItem("recent_files", "max_num_recent_files", temp);
  temp = selColor[0];
  conf.setItem("preferences", "selcolor_red", temp);
  temp = selColor[1];
  conf.setItem("preferences", "selcolor_green", temp);
  temp = selColor[2];
  conf.setItem("preferences", "selcolor_blue", temp);
  if(!conf.writeToFile(getConfigFile())) {
    std::cerr << "Writing configuration file failed" << std::endl;
  }
}
