// This file may be redistributed and modified only under the terms of
// the GNU General Public License (See COPYING for details).
// Copyright (C) 2003 Henri Kuuste

#include "model.h"
#include "image_loader.h"
#include "general.h"

Model::Model() :
  scale(1.0f),
  rootDir(""),
  calCoreModel(""),
  calModel(NULL),
  light(true),
  wireframe(false),
  surface(false),
  blenderLocked(false),
  boneSelected(false), selectedBone(0)
{
  selColor[0] = selColor[1] = selColor[2] = 1.0;
  createTree();
  scaleAd_ = new Gtk::Adjustment(scale, 0.0001, 10.0, 0.001, 0.05, 0.0);
  scaleAd_->signal_value_changed().connect(sigc::mem_fun(*this, &Model::getScaleFromAd));
  dirEntry = new Gtk::Entry();
  dirEntry->set_text(rootDir);
}

Model::~Model() {
  std::cout << "Deleting model data... " << std::endl;
  if (calModel != NULL) {
    delete calModel;
  }
  // calCoreModel.destroy();
  GLuint *textures = new GLuint[textureIds_.size()];
  for(unsigned int i = 0; i < textureIds_.size(); i++) {
    textures[i] = textureIds_[i];
  }
  glDeleteTextures(textureIds_.size(), textures);
  if(textureIds_.size() > 0 && !glIsTexture(textureIds_[0])) {
    std::cerr << "Textures Deleted..." << std::endl;
  }
  delete [] textures;
  textureIds_.clear();
  tree.clear();
  for(unsigned int i = 0; i < animCycles.size(); i++) {
    delete animCycles[i];
  }
  animCycles.clear();
  textureDirs.clear();
  if(scaleAd_) {
    delete scaleAd_;
  }
  if(dirEntry) {
    delete dirEntry;
  }
}


// LOADING FUNCTIONS

bool Model::loadSkeleton(const std::string &file) {
  std::cout << "Loading skeleton '" << file << "'..." << std::endl;
  if(calCoreModel.loadCoreSkeleton(rootDir + file))
    return true;
  return false;
}

int Model::addAnimationCycle(const std::string &file) {
  std::cout << "Loading animation cycle'" << file << "'..." << std::endl;
  int id = calCoreModel.loadCoreAnimation(rootDir + file);
  Gtk::TreeModel::iterator iter = tree->append(cycles.children());
  Gtk::TreeModel::Row row = *iter;
  row[cols.text] = Glib::path_get_basename(file);
  row[cols.active] = false;
  row[cols.id] = id;
  row[cols.selected] = false;
  row[cols.type] = ANIMATION_CYCLE;
  AnimationCycle *data = new AnimationCycle(iter);
  data->blending_changed.connect(
                                 sigc::mem_fun(*this, &Model::on_cycle_blend_changed));
  animCycles.push_back(data);
  return id;
}

int Model::addAnimationAction(const std::string &file) {
  std::cout << "Loading animation action'" << file << "'..." << std::endl;
  int id = calCoreModel.loadCoreAnimation(rootDir + file);
  Gtk::TreeModel::iterator iter = tree->append(actions.children());
  Gtk::TreeModel::Row row = *iter;
  row[cols.text] = Glib::path_get_basename(file);
  row[cols.active] = true;
  row[cols.id] = id;
  row[cols.selected] = false;
  row[cols.type] = ANIMATION_ACTION;
  AnimationCycle *data = new AnimationCycle(iter);
  animCycles.push_back(data);
  return id;
}

int Model::addMesh(const std::string &file) {
  std::cout << "Loading mesh '" << file << "'..." << std::endl;
  int id = calCoreModel.loadCoreMesh(rootDir + file);
  if (id == -1) {
    CalError::printLastError();
    return 0; 
  }
  Gtk::TreeModel::iterator iter = tree->append(meshes.children());
  Gtk::TreeModel::Row row = *iter;
  row[cols.text] = Glib::path_get_basename(file);
  row[cols.active] = true;
  row[cols.id] = id;
  row[cols.selected] = false;
  row[cols.type] = MESH;
  for(int i = 0; i < calCoreModel.getCoreMesh(id)->getCoreSubmeshCount(); i++) {
    iter = tree->append(row.children());
    Gtk::TreeModel::Row cRow = *iter;
    std::ostringstream str;
    str << "Submesh " << i;
    cRow[cols.text] = str.str();
    cRow[cols.active] = true;
    cRow[cols.id] = i;
    cRow[cols.selected] = false;
    cRow[cols.type] = SUBMESH;
  }
  return id;
}

int Model::addMaterial(const std::string &file) {
  std::cout << "Loading material '" << file << "'..." << std::endl;
  int id = calCoreModel.loadCoreMaterial(rootDir + file);
  textureDirs.push_back(Glib::path_get_dirname(rootDir+file) + G_DIR_SEPARATOR_S);
  Gtk::TreeModel::iterator iter = tree->append(materials.children());
  Gtk::TreeModel::Row row = *iter;
  row[cols.text] = Glib::path_get_basename(file);
  row[cols.active] = true;
  row[cols.id] = id;
  row[cols.selected] = false;
  row[cols.type] = MATERIAL;
  return id;
}

GLuint Model::loadTexture(const std::string &filename) {
  std::string ext = filename.substr(filename.length()-3);
  std::cout << "Loading " << ext << " texture '" << filename << "'..." << std::endl;
  if(ext == "raw") {
    return loadRAW(filename);
  } else {
    GLuint ret = loadGTK(filename);
    if(!ret) {
      std::cerr << "Could not load texture " << filename << std::endl;
      return 0;
    }
    return ret;
  }
}


bool Model::loadCFG(const std::string &filename) {
  std::ifstream file;
  file.open(filename.c_str(), std::ios::in | std::ios::binary);
  if(!file) {
    std::cerr << "Failed to open model configuration file '" << filename << "'." << std::endl;
    return false;
  }

  rootDir = fixPathSlashes(Glib::path_get_dirname(filename)) + G_DIR_SEPARATOR_S;
  std::cout << "Dir: " << rootDir << std::endl;
  dirEntry->set_text(rootDir);

  calCoreModel = CalCoreModel("model");
  // if(!calCoreModel.create("model")) {
    // CalError::printLastError();
    // return false;
  // }

  int line;
  for(line = 1; ; line++) {
    std::string strBuffer;
    std::getline(file, strBuffer);

    if(file.eof()) break;
    if(!file) {
      std::cerr << "Error while reading from the model configuration file '" << filename << "'." << std::endl;
      return false;
    }

    std::string::size_type pos = strBuffer.find_first_not_of(" \t");

    if((pos == std::string::npos) || (strBuffer[pos] == '\n') || (strBuffer[pos] == '\r') || (strBuffer[pos] == 0)) continue;

    if(strBuffer[pos] == '#') continue;

    std::string strKey = strBuffer.substr(pos, strBuffer.find_first_of(" =\t\n\r", pos) - pos);
    pos += strKey.size();

    pos = strBuffer.find_first_not_of(" \t", pos);
    if((pos == std::string::npos) || (strBuffer[pos] != '=')) {
      std::cerr << filename << "(" << line << "): " << strBuffer << std::endl << "**Invalid syntax." << std::endl;
      return false;
    }

    pos = strBuffer.find_first_not_of(" \t", pos + 1);

    std::string strData = strBuffer.substr(pos, strBuffer.find_first_of("\n\r", pos) - pos);

    if(strKey == "scale") {
      std::istringstream s(strData);
      float t;
      s >> t;
      setScale(t);
      std::cout << "Scale: " << scale << std::endl;
    } else if(strKey == "path") {
      rootDir = fixPathSlashes(strData);
      dirEntry->set_text(rootDir);
    } else if(strKey == "skeleton") {
      if(!loadSkeleton(fixPathSlashes(strData))) {
        CalError::printLastError();
        return false;
      }
    } else if(strKey == "animation" || strKey == "animation_cycle") {
      if(addAnimationCycle(fixPathSlashes(strData)) == -1) {
        CalError::printLastError();
        return false;
      }
    } else if(strKey == "animation_action") {
      if(addAnimationAction(fixPathSlashes(strData)) == -1) {
        CalError::printLastError();
        return false;
      }
    } else if(strKey == "mesh") {
      std::cerr << "data is " << strData << std::endl;
      if(addMesh(fixPathSlashes(strData)) == -1) {
        CalError::printLastError();
        return false;
      }
    } else if(strKey == "material") {
      if(addMaterial(fixPathSlashes(strData)) == -1) {
        CalError::printLastError();
        return false;
      }
    } else {
      std::cerr << filename << "(" << line << "): " << strBuffer << std::endl << "**Invalid syntax." << std::endl;
      return false;
    }
  }
  file.close();
  initTextures();
  createMaterialSet();
  createModel();
  return true;
}

// INITIALIZATION FUNCTIONS

void Model::initTextures() {
  for(int materialId = 0; materialId < calCoreModel.getCoreMaterialCount(); materialId++) {
    CalCoreMaterial *tCoreMaterial;
    tCoreMaterial = calCoreModel.getCoreMaterial(materialId);

    int mapId;
    for(mapId = 0; mapId < tCoreMaterial->getMapCount(); mapId++) {
      std::string filename;
      filename = fixPathSlashes(tCoreMaterial->getMapFilename(mapId));
      GLuint * textureId = new GLuint(loadTexture(textureDirs[materialId] + filename));
      textureIds_.push_back(*textureId);
      tCoreMaterial->setMapUserData(mapId, (Cal::UserData)textureId);
    }
  }
}

void Model::createMaterialSet() {
  for(int materialId = 0; materialId < calCoreModel.getCoreMaterialCount(); materialId++)
    {
      calCoreModel.createCoreMaterialThread(materialId);
      calCoreModel.setCoreMaterialId(materialId, 0, materialId);
    }
}

void Model::createModel() {
  calModel = new CalModel(&calCoreModel);
  // if(!calModel.create(&calCoreModel)) {
    // CalError::printLastError();
    // return;
  // }
	
  int meshId;
  for(meshId = 0; meshId < calCoreModel.getCoreMeshCount(); meshId++) {
    calModel->attachMesh(meshId);
  }
  calModel->setMaterialSet(0);
}

void Model::createTree() {
  tree = Gtk::TreeStore::create(cols);
  tree->clear();
  Gtk::TreeModel::iterator iter = tree->append();
  root = *iter;
  root[cols.text] = "Model";
  root[cols.active] = true;
  root[cols.selected] = false;
  root[cols.type] = MODEL;
  iter = tree->append(root.children());
  meshes = *iter;
  meshes[cols.text] = "Meshes";
  meshes[cols.active] = true;
  meshes[cols.selected] = false;
  meshes[cols.type] = MESHES;
  iter = tree->append(root.children());
  animations = *iter;
  animations[cols.text] = "Animations";
  animations[cols.active] = true;
  animations[cols.selected] = false;
  animations[cols.type] = ANIMATIONS;
  iter = tree->append(animations.children());
  cycles = *iter;
  cycles[cols.text] = "Cycles";
  cycles[cols.active] = true;
  cycles[cols.selected] = false;
  cycles[cols.type] = CYCLES;
  iter = tree->append(animations.children());
  actions = *iter;
  actions[cols.text] = "Actions";
  actions[cols.active] = true;
  actions[cols.selected] = false;
  actions[cols.type] = ACTIONS;
  iter = tree->append(root.children());
  materials = *iter;
  materials[cols.text] = "Materials";
  materials[cols.active] = true;
  materials[cols.selected] = false;
  materials[cols.type] = MATERIALS;
}


// RENDERING FUNCTIONS

void Model::setMaterialColor(CalRenderer *tCalRenderer) {
  unsigned char meshColor[4];
  GLfloat materialColor[4];

  tCalRenderer->getAmbientColor(&meshColor[0]);
  materialColor[0] = meshColor[0] / 255.0f;
  materialColor[1] = meshColor[1] / 255.0f;
  materialColor[2] = meshColor[2] / 255.0f;
  materialColor[3] = meshColor[3] / 255.0f;
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, materialColor);

  tCalRenderer->getDiffuseColor(&meshColor[0]);
  materialColor[0] = meshColor[0] / 255.0f;
  materialColor[1] = meshColor[1] / 255.0f;
  materialColor[2] = meshColor[2] / 255.0f;
  materialColor[3] = 1.0f;
  glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, materialColor);

  if(!light || (wireframe && surface)) {
    glColor4fv(materialColor);
  }

  tCalRenderer->getSpecularColor(&meshColor[0]);
  materialColor[0] = meshColor[0] / 255.0f;
  materialColor[1] = meshColor[1] / 255.0f;
  materialColor[2] = meshColor[2] / 255.0f;
  materialColor[3] = meshColor[3] / 255.0f;
  glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, materialColor);
}

void Model::render(bool select) {
  if(meshes[cols.active]) {
    Gtk::TreeModel::Children children = meshes.children();
    for(Gtk::TreeModel::Children::iterator iter = children.begin(); iter != children.end(); ++iter) {
      Gtk::TreeModel::Row row = *iter;
      if(row[cols.active]) {
        renderMeshByIter(iter, select);
      }
    }
  }
}

void Model::renderMeshByIter(Gtk::TreeModel::Children::iterator pIter, bool select) {
  CalRenderer *tCalRenderer;
  tCalRenderer = calModel->getRenderer();
  if(!tCalRenderer->beginRendering()) return;
	
  glMatrixMode(GL_MODELVIEW);
  glPushAttrib(GL_ALL_ATTRIB_BITS);
  glEnableClientState(GL_VERTEX_ARRAY);
  glEnableClientState(GL_NORMAL_ARRAY);
  Gtk::TreeModel::Children children = (*pIter).children();
  if(select) {
    glLoadName((*pIter)[cols.id]);
  }
	
  for(Gtk::TreeModel::Children::iterator iter = children.begin(); iter != children.end(); ++iter) {
    if((*iter)[cols.active]) {
      if(tCalRenderer->selectMeshSubmesh((*pIter)[cols.id], (*iter)[cols.id])) {
        if(select) {
          glPushName((*iter)[cols.id]);
        }
        setMaterialColor(tCalRenderer);
        bool selected = (*pIter)[cols.selected] || (*iter)[cols.selected];
			
        float shininess;
        //			shininess = tCalRenderer->getShininess();
        shininess = 30.0f;
        glMaterialfv(GL_FRONT, GL_SHININESS, &shininess);

        static float meshVertices[30000][3];
        int vertexCount;
        vertexCount = tCalRenderer->getVertices(&meshVertices[0][0]);

        static float meshNormals[30000][3];
        tCalRenderer->getNormals(&meshNormals[0][0]);

        static float meshTextureCoordinates[30000][2];
        int textureCoordinateCount;
        textureCoordinateCount = tCalRenderer->getTextureCoordinates(0,
                                                                     &meshTextureCoordinates[0][0]);

        static CalIndex meshFaces[50000][3];
        int faceCount;
        faceCount = tCalRenderer->getFaces(&meshFaces[0][0]);

        glVertexPointer(3, GL_FLOAT, 0, &meshVertices[0][0]);
        glNormalPointer(GL_FLOAT, 0, &meshNormals[0][0]);
				
        if((tCalRenderer->getMapCount() > 0) && (textureCoordinateCount > 0) && light)
          {
            glEnable(GL_TEXTURE_2D);
            glEnableClientState(GL_TEXTURE_COORD_ARRAY);
            glEnable(GL_COLOR_MATERIAL);
            glBindTexture(GL_TEXTURE_2D, *(GLuint*)tCalRenderer->getMapUserData(0));
            glTexCoordPointer(2, GL_FLOAT, 0, &meshTextureCoordinates[0][0]);
            glColor3f(1.0f, 1.0f, 1.0f);
          }

        glPushAttrib(GL_ENABLE_BIT | GL_POLYGON_BIT);
        if(selected || (wireframe && surface)) {
          glEnable(GL_POLYGON_OFFSET_FILL);
          glPolygonOffset(1,1);				
        }
        if(!(wireframe && !surface)) {
          if(sizeof(CalIndex)==2) {
            glDrawElements(GL_TRIANGLES, faceCount * 3, GL_UNSIGNED_SHORT,
                           &meshFaces[0][0]);
          } else {
            glDrawElements(GL_TRIANGLES, faceCount * 3, GL_UNSIGNED_INT,
                           &meshFaces[0][0]);
          }
        }
        if(selected || wireframe) {
          glDisable(GL_POLYGON_OFFSET_FILL);
          glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
          if(selected || surface) {
            glDisable(GL_LIGHTING);
            glDisable(GL_TEXTURE_2D);
          }
          if(selected) {
            glColor4f(selColor[0], selColor[1], selColor[2], 1);
          }
          glLineWidth(1.0f);
          if(sizeof(CalIndex)==2) {
            glDrawElements(GL_TRIANGLES, faceCount * 3, GL_UNSIGNED_SHORT,
                           &meshFaces[0][0]);
          } else {
            glDrawElements(GL_TRIANGLES, faceCount * 3, GL_UNSIGNED_INT,
                           &meshFaces[0][0]);
          }
          if(light) {
            glEnable(GL_LIGHTING);
          }
        }
        glPopAttrib();
        if(select) {
          glPopName();
        }
      }
    }
  }
  glDisableClientState(GL_NORMAL_ARRAY);
  glDisableClientState(GL_VERTEX_ARRAY);
  tCalRenderer->endRendering();
  glPopAttrib();
}

void Model::renderSkeleton(bool select)
{
  glPushAttrib( GL_ENABLE_BIT );
  glDisable( GL_LIGHTING);
  if(!select) {
    float lines[1024][2][3];
    int nrLines;
    nrLines = calModel->getSkeleton()->getBoneLines(&lines[0][0][0]);

    glLineWidth(3.0f);
    glColor3f(0.8f, 0.24f, 0.05f);
    glBegin(GL_LINES);
    int currLine;
    for(currLine = 0; currLine < nrLines; currLine++)
      {
        glVertex3f(lines[currLine][0][0], lines[currLine][0][1], lines[currLine][0][2]);
        glVertex3f(lines[currLine][1][0], lines[currLine][1][1], lines[currLine][1][2]);
      }
    glEnd();
    glLineWidth(1.0f);
  }

  float points[1024][3];
  int nrPoints;
  nrPoints = calModel->getSkeleton()->getBonePoints(&points[0][0]);

  glPointSize(6.0f);
  int currPoint;
  for(currPoint = 0; currPoint < nrPoints; currPoint++)
    {
      glColor3f(1.0f, 1.0f, 1.0f);
      if(select) {
        glLoadName(currPoint);
      }
      if(boneSelected) {
        if(currPoint == selectedBone) {
          glColor3f(0.2, 0.6, 0.3);
        }
      }
      glBegin(GL_POINTS);
      glVertex3f(points[currPoint][0], points[currPoint][1], points[currPoint][2]);
      glEnd();
    }
  glPointSize(1.0f);
  glPopAttrib();
}

void Model::update(float deltaTime) {
  calModel->update(deltaTime);
}


// UI FUNCTIONS

void Model::setLodLevel(float lodLevel) {
  calModel->setLodLevel(lodLevel);
}

void Model::on_cycle_blend_changed(Gtk::TreeIter iter) {
  if(!blenderLocked) {
    blenderLocked = true;
    //JOHNFLUX - Delay was 0.0, now 3.0
    blendCycle((*iter)[cols.id], 0, 3.0, true);
    blenderLocked = false;
  }
}

void Model::blendCycle(int id, float blending, float delay, bool fromScale) {
  if(fromScale) {
    animCycles[id]->on_blending_changed();
  } else {
    animCycles[id]->setBlending(blending);
  }
  if(animations[cols.active] && cycles[cols.active] && (*animCycles[id]->iter)[cols.active]) {
    Gtk::TreeModel::Children children = cycles.children();
    CalMixer *mixer = calModel->getMixer();
    float unchangedBlendSum = 0;
    bool noOtherActiveAnimations = true;
    for(Gtk::TreeModel::Children::iterator iter = children.begin(); iter != children.end(); ++iter) {
      if((*iter)[cols.id] != id && (*iter)[cols.active]) {
        noOtherActiveAnimations = false;
        unchangedBlendSum += animCycles[(*iter)[cols.id]]->getBlending();
      }
    }
    if(noOtherActiveAnimations) {
      animCycles[id]->setBlending(1.0);
    }
    float factor = unchangedBlendSum ? (1 - animCycles[id]->getBlending())/unchangedBlendSum : 0;
    for(Gtk::TreeModel::Children::iterator iter = children.begin(); iter != children.end(); ++iter) {
      if((*iter)[cols.active]) {
        if((*iter)[cols.id] != id) {
          animCycles[(*iter)[cols.id]]->setBlending(
                                                    animCycles[(*iter)[cols.id]]->getBlending()*factor);
        }
        mixer->blendCycle((*iter)[cols.id], animCycles[(*iter)[cols.id]]->getBlending(), delay);
      }
    }
  }
}

void Model::disableCycle(int id) {
  Gtk::TreeModel::Children children = cycles.children();
  bool noOtherActiveAnimations = true;
  for(Gtk::TreeModel::Children::iterator iter = children.begin(); iter != children.end(); ++iter) {
    if((*iter)[cols.id] != id && (*iter)[cols.active]) {
      noOtherActiveAnimations = false;
      break;
    }
  }
  if(noOtherActiveAnimations) {
    animCycles[id]->setBlending(0);
  } else {
    (*animCycles[id]->iter)[cols.active] = true;
    //JOHNFLUX = delay was 0.5, now 3.0
    blendCycle(id, 0, 3.0, false);
    (*animCycles[id]->iter)[cols.active] = false;
  }
  calModel->getMixer()->clearCycle(id, 0.5);
}

void Model::enableCycle(int id) {
  //JOHNFLUX - delay was 0.5, now 3.0
  blendCycle(id, animCycles[id]->getBlending(), 3.0);
}

void Model::executeAction(int id, float in, float out) {
  CalMixer *mixer = calModel->getMixer();
  mixer->executeAction(id, in, out);
}

void Model::setMaterial(int meshId, int submeshId, int materialId) {
  calModel->getMesh(meshId)->getSubmesh(submeshId)->setCoreMaterialId(materialId);
}

int Model::getPolyCount() {
  int polyCount = 0;
  std::vector<CalMesh*> meshes = calModel->getVectorMesh();
  for(unsigned int i = 0; i < meshes.size(); i++) {
    std::vector<CalSubmesh*> submeshes = meshes[i]->getVectorSubmesh();
    for(unsigned int j = 0; j < submeshes.size(); j++) {
      polyCount += submeshes[j]->getFaceCount();
    }
  }
  return polyCount;
}

int Model::getMeshPolyCount(int id) {
  int polyCount = 0;
  std::vector<CalMesh*> meshes = calModel->getVectorMesh();
  if(meshes[id]) {
    std::vector<CalSubmesh*> submeshes = meshes[id]->getVectorSubmesh();
    for(unsigned int i = 0; i < submeshes.size(); i++) {
      polyCount += submeshes[i]->getFaceCount();
    }
  }
  return polyCount;
}

int Model::getSubmeshPolyCount(int meshId, int submeshId) {
  std::vector<CalMesh*> meshes = calModel->getVectorMesh();
  if(meshes[meshId]) {
    std::vector<CalSubmesh*> submeshes = meshes[meshId]->getVectorSubmesh();
    if(submeshes[submeshId]) {
      return submeshes[submeshId]->getFaceCount();
    }
  }
  return 0;
}

void Model::getSelColor(double *color) {
  color[0] = selColor[0];
  color[1] = selColor[1];
  color[2] = selColor[2];
}

void Model::setSelColor(double *color) {
  selColor[0] = color[0];
  selColor[1] = color[1];
  selColor[2] = color[2];
}

void Model::getScaleFromAd() {
  scale = scaleAd_->get_value();
}

void Model::setScale(float sc) {
  if(sc < 0.0001) {
    sc = 0.0001;
  }
  if(sc > 10.0) {
    sc = 10.0;
  }
  scale = sc;
  scaleAd_->set_value(sc);
}

Gtk::SpinButton *Model::createScaleSpin() {
  Gtk::SpinButton *t = new Gtk::SpinButton(*scaleAd_, 1.0, 4);
  t->set_numeric(true);
  return t;
}
