// This file may be redistributed and modified only under the terms of
// the GNU General Public License (See COPYING for details).
// Copyright (C) 2003 Henri Kuuste
// Changes Copyright 2004 Kai Blin

#ifndef _ECMODEL_H_
#define _ECMODEL_H_

#include <cal3d/cal3d.h>
#include <string>
#include <GL/gl.h>
#include <GL/glu.h>

#include <cstdio>
#include <vector>

#include <gtkmm.h>
#include <glibmm.h>

#include <sigc++/object.h>
#include <sigc++/object_slot.h>

#include <fstream>
#include <sstream>
#include "model_cols.h"
#include "animation_cycle.h"

/**
 * Base class for cal3d models
 *
 *\author Henri Kuuste
 *\author Kai Blin (documentation)
 *
 */

class Model : public sigc::trackable {

public:

  /**
   * Default constructor
   */
  Model();

  /**
   * Default destructor
   */
  ~Model();


  /**
   * Load a model .cfg file and parse it, loading the meshes/skeleton/textures 
   * as they come.
   *
   *\param filename std::string containing the filename
   *\return true on success, false otherwise
   */
  bool loadCFG(const std::string &filename);

  /**
   * Load a cal3d skeleton from a file
   *
   *\param file std::string containing filename to load
   *\return true if it worked, false otherwise
   */
  bool loadSkeleton(const std::string &file);

  /**
   * Add an animation cycle from a file, adding it to the GTK Treeview
   *
   *\param file std::string containing filename to load
   *\return the result of the cal3d loadCoreAnimation call
   */
  int addAnimationCycle(const std::string &file);

  /**
   * Add a one time (not looping) animation from a file, adding it to the 
   * GTK Treeview
   *
   *\param file std::string containing filename to load
   *\return the result of the cal3d loadCoreAnimation call
   */
  int addAnimationAction(const std::string &file);

  /**
   * Add a cal3d mesh from a file, adding it and all submeshes to the GTK
   * Treeview
   *
   *\param file std::string containing the filename
   *\return the result of the cal3d loadCoreMesh call, or 0 if it failed
   */
  int addMesh(const std::string &file);

  /**
   * Add a material from a file, adding it to the GTK Treeview
   *
   *\param file std::string containing the filename
   *\return the result of the cal3d loadCoreMaterial call
   */
  int addMaterial(const std::string &file);

  /**
   * For every material, set up a cal3d core material thread
   */
  void createMaterialSet();

  /**
   * Create the cal3d model, loading all the meshes and setting the default material
   */
  void createModel();
  
  /**
   * Create the GTK Tree with all the subnodes and leaves needed
   */
  void createTree();

  /**
   * Set up the OpenGL colors using the colors given by the CalRenderer
   *
   *\param tCalRenderer pointer to curren CalRenderer 
   */
  void setMaterialColor(CalRenderer *tCalRenderer);

  /**
   * Load a texture from a file, picking the right loading function
   * based on the extension
   *
   * It uses the loadRAW(std::string) function if the file is in .raw format,
   * or the loadGTK(std::string) function otherwise
   *
   *\param file std::string containing the filename
   *\return the result of the corresponding load call, or 0 if it failed
   */
  GLuint loadTexture(const std::string &filename);

  /**
   * Set up the textures as materials
   */
  void initTextures();

  /**
   * Render all selected meshes
   *
   *\param select is the mesh selected? Defaults to false
   */
  void render(bool select = false);

  /**
   * Iterate over the meshes and render them if they are selected
   *
   *\param pIter iterator over the TreeModel mesh subtree
   *\param select is the mesh selected? Defaults to false
   */
  void renderMeshByIter(Gtk::TreeModel::Children::iterator pIter, bool select = false);

  /**
   * Render the skeleton, and the color of the selected bone if there is one
   *
   *\param select is there a selected bone? Defaults to false
   */
  void renderSkeleton(bool select = false);

  /**
   * Update the model
   *
   *\param deltaTime some time difference
   *\todo look this one up
   */
  void update(float deltaTime);

  /**
   * Set the level of detail
   *
   *\param lodLevel level to set the LOD to
   */
  void setLodLevel(float lodLevel);

  /**
   * Callback to handle changes to the blend cycle
   *
   * John added a delay of 3.0 here
   *
   *\param iter pointer to the blending cycle that changed
   */
  void on_cycle_blend_changed(Gtk::TreeIter iter);

  /**
   * Blend in an animation cycle
   *
   *\param id the ID of the cycle to render
   *\param blending how much of the current animation to blend in
   *\param delay how long to wait until finishing to blend in the animation
   *\param fromScale should we get the blending ratio from the GUI scale? Defaults to false
   */
  void blendCycle(int id, float blending, float delay, bool fromScale = false);

  /**
   * Turn off an animation cylcle
   *
   * John increased the delay here
   *
   *\param id ID of the cycle to turn off
   */
  void disableCycle(int id);

  /**
   * Switch on an animation cycle
   *
   * John increased the delay here
   *\param id ID of the cycle to switch on 
   */
  void enableCycle(int id);

  /**
   *
   */
  void setBlendWeight(int id, float weight);

  /**
   * Execute a one-shot animation
   * 
   *\param id ID of extion to execute
   *\param in I have no clue what it does
   *\param out I have no clue what it does
   *\todo fix the documentation of this function
   */
  void executeAction(int id, float in, float out);

  /**
   * Set the material of a submesh
   *
   *\param meshId ID of the mesh to take the submesh of
   *\param submeshId ID of the submesh to set the material
   *\param materialId ID of the material to set
   */
  void setMaterial(int meshId, int submeshId, int materialId);

  /**
   * Get the scale from the GTK::Adjustment
   */
  void getScaleFromAd();

  /**
   * Set the cale and cap it between 0.0001 and 10.0
   */
  void setScale(float sc);

  /**
   * Get the polygon count
   *
   *\return the number of polygons in the model
   */
  int getPolyCount();

  /**
   * Get the polygon count of a mesh
   *
   *\param id ID of the mesh to count the polygons in
   *\return the number of polygons in the selected mesh
   */
  int getMeshPolyCount(int id);

  /**
   * Get the polygon count of a submesh
   *
   *\param id ID of the submesh to count the polygons in
   *\return the number of polygons in the selected submesh
   */
  int getSubmeshPolyCount(int meshId, int submeshId);

  /**
   * Get the selected color and store it in an array
   *
   *\param color array of double to store the color in
   */
  void getSelColor(double *color);

  /**
   * Set the selected color
   *
   *\param color array of double to set the color from
   */
  void setSelColor(double *color);

  /**
   * Scale of the model, defaults to 1.0
   */
  float                    scale;

  /**
   * Root directory to load files from
   */
  std::string              rootDir;

  /**
   * A std::vector containing the texture directories
   */
  std::vector<std::string> textureDirs;

  /**
   * The cal3d core model
   */
  CalCoreModel             calCoreModel;

  /**
   * The cal3d model
   */
  CalModel *               calModel;

  /**
   * Is the light source switched on? Defaults to true
   */
  bool                     light;

  /**
   * Is wireframe rendering switched on? Defaults to false
   */
  bool                     wireframe;

  /**
   * Is there a surface to render? Defaults to false
   */
  bool                     surface;

  /**
   * Is the cycle blender currently locked? Defaults to false
   */
  bool                     blenderLocked;

  /**
   * Is there a bone selected? Defaults to false
   */
  bool                     boneSelected;

  /**
   * The ID of the selected bone, defaults to 0
   */
  int                      selectedBone;

  /**
   * Create a spinbutton set to the current scale
   */
  Gtk::SpinButton          *createScaleSpin();

  /**
   * A GTK::Entry, set to the root dir
   */
  Gtk::Entry               *dirEntry;

  /**
   * I have no clue what this is
   *\todo find out what this is
   */ 
  Glib::RefPtr<Gtk::TreeStore> tree;
	
  /**
   * The different subtrees of the GTK tree model
   */
  Gtk::TreeModel::Row root, meshes, animations, materials, cycles, actions;

  /**
   * A std::vector containing all the animation cycles
   */
  std::vector<AnimationCycle*>	animCycles;
	
protected:
  double                selColor[3];
  Gtk::Adjustment       *scaleAd_;
  std::vector<GLuint>   textureIds_;
  ModelColumns          cols;

};

#endif
